from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail


class APIException(Exception):
    def __init__(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
        exception: Exception,
    ) -> None:
        msg = f"{exchange.name}: Error fetching market data for market '{market}' and resolution '{resolution.notation}: {exception}'."
        super().__init__(msg)



from dataclasses import dataclass
from abc import ABC
from enum import IntEnum


class Seconds(IntEnum):
    ONE_WEEK = 60 * 60 * 24 * 7
    ONE_DAY = 60 * 60 * 24
    FOUR_HOURS = 60 * 60 * 4
    ONE_HOUR = 60 * 60
    THIRTY_MINUTES = 60 * 30
    FIFTEEN_MINUTES = 60 * 15
    FIVE_MINUTES = 60 * 5
    ONE_MINUTE = 60


@dataclass
class ResolutionDetail:
    seconds: Seconds
    notation: str


class ExchangeResolution(ABC):
    @classmethod
    def notation_to_resolution_detail(cls, notation: str) -> ResolutionDetail:
        seconds = cls._notation_to_seconds(notation)
        return ResolutionDetail(seconds, notation)

    @classmethod
    def _notation_to_seconds(cls, notation: str) -> Seconds:
        for attr in dir(cls):
            if (
                isinstance(getattr(cls, attr), ResolutionDetail)
                and getattr(cls, attr).notation == notation
            ):
                return getattr(cls, attr).seconds
        raise ValueError(f"Invalid notation: {notation}.")



from enum import Enum
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle.exchange import Exchange


class PricePoint(Enum):
    OPEN = "open"
    HIGH = "high"
    LOW = "low"
    CLOSE = "close"


@dataclass
class Candle:
    id: int | None
    exchange: Exchange
    market: str
    resolution: ResolutionDetail

    started_at: datetime

    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal

    base_token_volume: Decimal
    trades: int
    usd_volume: Decimal
    starting_open_interest: Decimal



from decimal import Decimal
from typing import Any, Dict
from locast.candle.candle import Candle
from locast.candle.exchange_candle_mapping import ExchangeCandleMapping
from locast.candle.dydx.dydx_resolution import DydxResolution


from sir_utilities.date_time import string_to_datetime

from locast.candle.exchange import Exchange


class DydxV4CandleMapping(ExchangeCandleMapping):
    def to_candle(self, exchange_candle: Dict[str, Any]) -> Candle:
        try:
            started_at = string_to_datetime(exchange_candle["startedAt"])
            market = exchange_candle["ticker"]
            resolution = DydxResolution.notation_to_resolution_detail(
                exchange_candle["resolution"]
            )
            p_low = Decimal(exchange_candle["low"])
            p_high = Decimal(exchange_candle["high"])
            p_open = Decimal(exchange_candle["open"])
            p_close = Decimal(exchange_candle["close"])
            base_token_volume = Decimal(exchange_candle["baseTokenVolume"])
            usd_volume = Decimal(exchange_candle["usdVolume"])
            trades = int(exchange_candle["trades"])
            starting_open_interest = Decimal(exchange_candle["startingOpenInterest"])

            return Candle(
                id=None,
                exchange=Exchange.DYDX_V4,
                started_at=started_at,
                market=market,
                resolution=resolution,
                low=p_low,
                high=p_high,
                open=p_open,
                close=p_close,
                base_token_volume=base_token_volume,
                usd_volume=usd_volume,
                trades=trades,
                starting_open_interest=starting_open_interest,
            )
        except Exception as e:
            raise ValueError(f"Error converting dict to Candle: {str(e)}") from e



from typing import Protocol, Any


from locast.candle.candle import Candle


class DatabaseCandleMapping(Protocol):
    def to_candle(self, database_candle: Any) -> Candle: ...

    def to_database_candle(self, candle: Candle) -> Any: ...



from sqlmodel import (
    Session,
    select,  # type: ignore
)
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_storage.sql.tables import (
    SqliteExchange,
    SqliteMarket,
    SqliteResolution,
)


class TableUtility:
    @staticmethod
    def lookup_sqlite_exchange(
        exchange: Exchange,
        session: Session,
    ) -> SqliteExchange | None:
        return session.exec(select(SqliteExchange).filter_by(exchange=exchange)).first()

    @staticmethod
    def lookup_sqlite_market(market: str, session: Session) -> SqliteMarket | None:
        return session.exec(select(SqliteMarket).filter_by(market=market)).first()

    @staticmethod
    def lookup_sqlite_resolution(
        resolution: ResolutionDetail,
        session: Session,
    ) -> SqliteResolution | None:
        return session.exec(
            select(SqliteResolution).filter_by(
                seconds=resolution.seconds,
                notation=resolution.notation,
            )
        ).first()

    @staticmethod
    def lookup_or_insert_sqlite_exchange(
        exchange: Exchange,
        session: Session,
    ) -> SqliteExchange:
        sql_exchange = TableUtility.lookup_sqlite_exchange(
            exchange=exchange,
            session=session,
        )

        if not sql_exchange:
            sql_exchange = SqliteExchange(exchange=exchange)
            session.add(sql_exchange)
            session.commit()
        return sql_exchange

    @staticmethod
    def lookup_or_insert_sqlite_market(
        market: str,
        session: Session,
    ) -> SqliteMarket:
        sql_market = TableUtility.lookup_sqlite_market(market=market, session=session)

        if not sql_market:
            sql_market = SqliteMarket(market=market)
            session.add(sql_market)
            session.commit()
        return sql_market

    @staticmethod
    def lookup_or_insert_sqlite_resolution(
        resolution: ResolutionDetail,
        session: Session,
    ) -> SqliteResolution:
        sql_resolution = TableUtility.lookup_sqlite_resolution(
            resolution=resolution,
            session=session,
        )

        if not sql_resolution:
            sql_resolution = SqliteResolution(
                seconds=resolution.seconds,
                notation=resolution.notation,
            )
            session.add(sql_resolution)
            session.commit()
        return sql_resolution



from typing import List, Dict, Any
from locast.candle.candle import Candle
from locast.candle.exchange_candle_mapping import ExchangeCandleMapping


class ExchangeCandleMapper:
    """
    A class that maps candle dictionaries from different exchanges to Candle objects.
    Note: _to_candle() exists, to extract the _select_mapping() call outside of the for loop
    in to_candles(), in order to only have it run every batch instead of every candle.
    """

    def __init__(self, mapping: ExchangeCandleMapping) -> None:
        self._mapping = mapping

    def to_candles(self, candle_dicts: List[Dict[str, Any]]) -> List[Candle]:
        return [
            self._to_candle(self._mapping, candle_dict) for candle_dict in candle_dicts
        ]

    def to_candle(self, candle_dict: Dict[str, Any]) -> Candle:
        return self._to_candle(self._mapping, candle_dict)

    def _to_candle(
        self,
        mapping: ExchangeCandleMapping,
        candle_dict: Dict[str, Any],
    ) -> Candle:
        return mapping.to_candle(candle_dict)



import asyncio

from dydx_v4_client.indexer.rest.indexer_client import IndexerClient  # type: ignore
from dydx_v4_client.network import make_mainnet  # type: ignore
from locast.candle.dydx.dydx_resolution import DydxResolution
from locast.candle_fetcher.dydx.candle_fetcher.dydx_candle_fetcher import (
    DydxCandleFetcher,
)
from locast.candle_fetcher.dydx.api_fetcher.dydx_v4_fetcher import DydxV4Fetcher
from locast.candle.candle_utility import CandleUtility as cu

MAINNET = make_mainnet(  # type: ignore
    rest_indexer="https://indexer.dydx.trade/",
    websocket_indexer="wss://indexer.dydx.trade/v4/ws",
    node_url="dydx-ops-rpc.kingnodes.com:443",
)


async def main() -> None:
    res = DydxResolution.FIFTEEN_MINUTES
    market = "DOGE-USD"
    dydx_v4_fetcher = DydxV4Fetcher(IndexerClient(MAINNET.rest_indexer))
    fetcher = DydxCandleFetcher(dydx_v4_fetcher)

    horizon = await fetcher.find_horizon(market, res)
    print(f"horizon: {horizon}")

    horizon_plus_one = cu.add_one_resolution(horizon, res)
    horizon_minus_one = cu.subtract_n_resolutions(horizon, res, 1)
    horizon_minus_two = cu.subtract_n_resolutions(horizon, res, 2)
    oldest_candle = await fetcher.fetch_candles(
        market,
        res,
        horizon,
        horizon_plus_one,
    )

    one_older = await fetcher.fetch_candles(
        market,
        res,
        horizon_minus_one,
        horizon,
    )

    two_older = await fetcher.fetch_candles(
        market,
        res,
        horizon_minus_two,
        horizon_minus_one,
    )

    print("Oldest:")
    print(oldest_candle)
    print("One older:")
    print(one_older)
    print("Two older:")
    print(two_older)


asyncio.run(main())



from typing import Any

from locast.candle.candle import Candle
from locast.candle_storage.database_candle_mapping import DatabaseCandleMapping


class DatabaseCandleMapper:
    """
    Maps candle representations from different database technologies to Candle objects and vice versa.
    """

    def __init__(self, mapping: DatabaseCandleMapping) -> None:
        self._mapping = mapping

    def to_candle(self, database_candle: Any) -> Candle:
        return self._mapping.to_candle(database_candle)

    def to_database_candle(self, candle: Candle) -> Any:
        return self._mapping.to_database_candle(candle)



from datetime import datetime
from sqlalchemy import Index, UniqueConstraint
from sqlmodel import Field, Relationship, SQLModel  # type: ignore

from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import Seconds


class SqliteExchange(SQLModel, table=True):
    __tablename__ = "exchange"  # type: ignore
    id: int = Field(default=None, primary_key=True)
    exchange: Exchange = Field(nullable=False, unique=True)


class SqliteResolution(SQLModel, table=True):
    __tablename__ = "resolution"  # type: ignore
    id: int = Field(default=None, primary_key=True)
    seconds: Seconds = Field(nullable=False, unique=True)
    notation: str = Field(nullable=False, unique=True)


class SqliteMarket(SQLModel, table=True):
    __tablename__ = "market"  # type: ignore
    id: int = Field(default=None, primary_key=True)
    market: str = Field(nullable=False, unique=True)


# NOTE: For id to be NULL in db (which makes no sense) you could set nullable=True.
class SqliteCandle(SQLModel, table=True):
    __tablename__ = "candle"  # type: ignore
    id: int | None = Field(default=None, primary_key=True)
    exchange_id: int = Field(
        default=None,
        foreign_key=("exchange.id"),
        nullable=False,
        index=True,
    )
    market_id: int = Field(
        default=None,
        foreign_key=("market.id"),
        nullable=False,
        index=True,
    )
    resolution_id: int = Field(
        default=None,
        foreign_key=("resolution.id"),
        nullable=False,
        index=True,
    )

    started_at: datetime

    open: str
    high: str
    low: str
    close: str

    base_token_volume: str
    trades: int
    usd_volume: str
    starting_open_interest: str

    # Relationships
    exchange: SqliteExchange = Relationship()
    market: SqliteMarket = Relationship()
    resolution: SqliteResolution = Relationship()

    __table_args__ = (
        Index(
            "compound_index_exchange_market_resolution",
            "exchange_id",
            "market_id",
            "resolution_id",
        ),
        UniqueConstraint(
            "exchange_id",
            "market_id",
            "resolution_id",
            "started_at",
            name="unique_candle_constraint",
        ),
    )



from datetime import datetime
from typing import Protocol, List, runtime_checkable

from locast.candle.candle import Candle
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail


@runtime_checkable
class CandleFetcher(Protocol):
    @property
    def exchange(self) -> Exchange: ...

    @property
    def log_progress(self) -> bool: ...

    @log_progress.setter
    def log_progress(self, value: bool) -> None: ...

    async def fetch_candles(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
        end_date: datetime,
    ) -> List[Candle]: ...

    async def fetch_candles_up_to_now(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
    ) -> List[Candle]: ...

    async def find_horizon(
        self,
        market: str,
        resolution: ResolutionDetail,
    ) -> datetime: ...



from datetime import datetime
from typing import Dict, List

from locast.candle.candle_utility import CandleUtility as cu
from locast.candle.candle import Candle
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_fetcher.candle_fetcher import CandleFetcher
from locast.candle_storage.cluster_info import ClusterInfo
from locast.candle_storage.candle_storage import CandleStorage
from locast.logging_functions import (
    log_redundant_call,
    log_start_date_shifted_to_horizon,
)

# TODO: Expand API by adding convenience methods such as def retrieve_segment(from, to, exchange, market, resolution) -> List[Candle]: ...


class StoreManager:
    def __init__(
        self,
        candle_fetcher: CandleFetcher,
        candle_storage: CandleStorage,
    ) -> None:
        self._candle_fetcher = candle_fetcher
        self._candle_storage = candle_storage

        self._horizon_cache: Dict[str, datetime] = {}

    async def create_cluster(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
        replace_existing_cluster: bool = False,
    ) -> None:
        # 1) Check if cluster exists in candle store
        cluster_info = await self.get_cluster_info(
            self._candle_fetcher.exchange,
            market,
            resolution,
        )

        if cluster_info.newest_candle:
            if not replace_existing_cluster:
                raise ExistingClusterException(
                    f"Cluster already exists for market {market} and resolution {resolution.notation}."
                )

            await self.delete_cluster(
                self._candle_fetcher.exchange,
                market,
                resolution,
            )

        start_date = await self._check_horizon(market, resolution, start_date)

        cluster = await self._candle_fetcher.fetch_candles_up_to_now(
            market,
            resolution,
            start_date,
        )

        await self._candle_storage.store_candles(cluster)

    async def retrieve_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> List[Candle]:
        cluster_info = await self.get_cluster_info(exchange, market, resolution)

        if cluster_info.size == 0:
            raise MissingClusterException(
                f"Cluster does not exist for market {market} and resolution {resolution.notation}."
            )

        return await self._candle_storage.retrieve_cluster(exchange, market, resolution)

    async def retrieve_newest_candles(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
        amount: int,
    ) -> List[Candle]:
        cluster_info = await self.get_cluster_info(exchange, market, resolution)

        if cluster_info.size == 0:
            raise MissingClusterException(
                f"Cluster does not exist for market {market} and resolution {resolution.notation}."
            )

        amount_to_retrieve = min(amount, cluster_info.size)

        return await self._candle_storage.retrieve_newest_candles(
            exchange,
            market,
            resolution,
            amount_to_retrieve,
        )

    async def update_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> None:
        cluster_info = await self.get_cluster_info(exchange, market, resolution)

        if cluster_info.is_uptodate:
            next_tick = cu.next_tick(resolution)
            msg = f"Cluster is already up to date. Try again after {next_tick}."
            log_redundant_call("♻️", msg)
            return

        if not (head := cluster_info.newest_candle):
            raise MissingClusterException(
                f"Cluster does not exist for market {market} and resolution {resolution.notation}."
            )

        start_date = cu.add_one_resolution(head.started_at, resolution)
        new_candles = await self._candle_fetcher.fetch_candles_up_to_now(
            market,
            resolution,
            start_date,
        )

        await self._candle_storage.store_candles(new_candles)

    async def delete_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> None:
        cluster_info = await self.get_cluster_info(exchange, market, resolution)

        if not cluster_info.newest_candle:
            raise MissingClusterException(
                f"Cluster does not exist for market {market} and resolution {resolution.notation}."
            )

        await self._candle_storage.delete_cluster(
            exchange,
            market,
            resolution,
        )

    async def get_cluster_info(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> ClusterInfo:
        return await self._candle_storage.get_cluster_info(exchange, market, resolution)

    async def _check_horizon(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
    ) -> datetime:
        if not (horizon := self._horizon_cache.get(f"{market}_{resolution.notation}")):
            horizon = await self._candle_fetcher.find_horizon(market, resolution)
            self._horizon_cache[f"{market}_{resolution.notation}"] = horizon

        if start_date < horizon:
            log_start_date_shifted_to_horizon(
                "⏭️",
                self._candle_fetcher.exchange,
                start_date,
                horizon,
            )
            start_date = horizon
        return start_date


class ExistingClusterException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class MissingClusterException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)



from decimal import Decimal
from sqlmodel import Session
from datetime import timezone

from locast.candle.candle import Candle
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_storage.database_candle_mapping import DatabaseCandleMapping
from locast.candle_storage.sql.tables import (
    SqliteCandle,
    SqliteExchange,
    SqliteMarket,
    SqliteResolution,
)

from locast.candle_storage.sql.table_utility import TableUtility as tu


class SqliteCandleMapping(DatabaseCandleMapping):
    def __init__(self, session: Session | None = None) -> None:
        self._session = session

        self._sql_exchange_cache: SqliteExchange | None = None
        self._sql_market_cache: SqliteMarket | None = None
        self._sql_resolution_cache: SqliteResolution | None = None

        self._exchange_cache: Exchange | None = None
        self._market_cache: str | None = None
        self._resolution_cache: ResolutionDetail | None = None

    def to_candle(self, database_candle: SqliteCandle) -> Candle:
        try:
            if not self._exchange_cache:
                self._exchange_cache = database_candle.exchange.exchange
            if not self._market_cache:
                self._market_cache = database_candle.market.market
            if not self._resolution_cache:
                self._resolution_cache = ResolutionDetail(
                    database_candle.resolution.seconds,
                    database_candle.resolution.notation,
                )

            return Candle(
                id=database_candle.id,
                exchange=self._exchange_cache,
                market=self._market_cache,
                resolution=self._resolution_cache,
                started_at=database_candle.started_at.replace(tzinfo=timezone.utc),
                open=Decimal(database_candle.open),
                high=Decimal(database_candle.high),
                low=Decimal(database_candle.low),
                close=Decimal(database_candle.close),
                base_token_volume=Decimal(database_candle.base_token_volume),
                trades=database_candle.trades,
                usd_volume=Decimal(database_candle.usd_volume),
                starting_open_interest=Decimal(database_candle.starting_open_interest),
            )
        except Exception as e:
            raise e

    # NOTE: lookup_or_insert funcs are only ever used here, before WRITING to database
    def to_database_candle(self, candle: Candle) -> SqliteCandle:
        try:
            assert (
                self._session
            ), f"Session must be provided to map to {SqliteCandle.__name__}."
            with self._session as session:
                if not self._sql_exchange_cache:
                    self._sql_exchange_cache = tu.lookup_or_insert_sqlite_exchange(
                        candle.exchange,
                        session,
                    )
                if not self._sql_market_cache:
                    self._sql_market_cache = tu.lookup_or_insert_sqlite_market(
                        candle.market,
                        session,
                    )
                if not self._sql_resolution_cache:
                    self._sql_resolution_cache = tu.lookup_or_insert_sqlite_resolution(
                        candle.resolution,
                        session,
                    )

                return SqliteCandle(
                    id=candle.id,
                    exchange=self._sql_exchange_cache,
                    exchange_id=self._sql_exchange_cache.id,  # needed for bulk
                    market=self._sql_market_cache,
                    market_id=self._sql_market_cache.id,  # needed for bulk
                    resolution=self._sql_resolution_cache,
                    resolution_id=self._sql_resolution_cache.id,  # needed for bulk
                    started_at=candle.started_at,
                    open=str(candle.open),
                    high=str(candle.high),
                    low=str(candle.low),
                    close=str(candle.close),
                    base_token_volume=str(candle.base_token_volume),
                    trades=candle.trades,
                    usd_volume=str(candle.usd_volume),
                    starting_open_interest=str(candle.starting_open_interest),
                )
        except Exception as e:
            raise e



from enum import Enum
from typing import List, TypeVar
from zoneinfo import ZoneInfo
from datetime import datetime, timedelta, timezone
from locast.candle.candle import Candle
from locast.candle.exchange_resolution import ResolutionDetail

EnumType = TypeVar("EnumType", bound=Enum)

# TODO: Check what functions are not needed anymore.


class CandleUtility:
    @classmethod
    def is_newest_valid_candle(
        cls,
        candle: Candle,
    ) -> bool:
        # valid_up_to is the date a newest candle should have as started_at in order
        # to be valid - if candle.started_at differs from it, the candle is either
        # unfinished (its date is newer than valid_up_to) or not the newest, (its date
        # lacks behind) meaning we have a gap to fill.
        valid_up_to = cls.valid_up_to(candle.resolution)
        started_at = candle.started_at.replace(tzinfo=ZoneInfo("UTC"))

        return started_at == valid_up_to

    @classmethod
    def valid_up_to(cls, resolution: ResolutionDetail) -> datetime:
        # check the most recent date at which candles for this resolution are valid
        last_tick = cls.norm_date(datetime.now(timezone.utc), resolution)
        return (last_tick - timedelta(seconds=resolution.seconds)).replace(
            tzinfo=ZoneInfo("UTC")
        )

    # @classmethod
    # def last_tick(cls, resolution: ResolutionDetail) -> datetime:
    #     # calculate the last tick date of that resolution
    #     utc_now = datetime.now(timezone.utc)
    #     unix_epoch = datetime.fromtimestamp(0, timezone.utc)
    #     now_seconds = (utc_now - unix_epoch).total_seconds()

    #     remainder_sec = now_seconds % resolution.seconds
    #     last_tick_sec = now_seconds - remainder_sec
    #     return datetime.fromtimestamp(last_tick_sec, ZoneInfo("UTC"))

    @classmethod
    def next_tick(cls, resolution: ResolutionDetail) -> datetime:
        utc_now = cls.norm_date(datetime.now(timezone.utc), resolution)
        return utc_now + timedelta(
            seconds=resolution.seconds - (utc_now.second % resolution.seconds)
        )

    @classmethod
    def norm_date(cls, date: datetime, resolution: ResolutionDetail) -> datetime:
        """
        Normalize a given datetime to the nearest lower multiple of a given resolution.
        In other words: Round down a date by a given resolution.
        Examples:
        12:37:12 rounded down by 1min resolution becomes 12:37
        12:37:12 rounded down by 5min resolution becomes 12:35
        12:37:12 rounded down by 15min resolution becomes 12:30


        Args:
            date (datetime): The datetime to be normalized.
            resolution (ResolutionDetail): The resolution to which the datetime should be normalized.


        Returns:
            datetime: The normalized datetime.
        """
        # Calculate the number of seconds since the epoch
        seconds_since_epoch = (
            date - datetime.fromtimestamp(0, tz=date.tzinfo)
        ).total_seconds()

        # Calculate the normalized seconds
        normalized_seconds = seconds_since_epoch - (
            seconds_since_epoch % resolution.seconds
        )

        # Create a new datetime with the normalized seconds
        return datetime.fromtimestamp(0, tz=date.tzinfo) + timedelta(
            seconds=normalized_seconds
        )

    @classmethod
    def normalized_now(cls, resolution: ResolutionDetail) -> datetime:
        # Generate the startedAt date for the newest finished candle (now - 1 res)
        return cls.norm_date(datetime.now(timezone.utc), resolution)

    @classmethod
    def subtract_n_resolutions(
        cls,
        date: datetime,
        resolution: ResolutionDetail,
        n: int,
    ) -> datetime:
        return date - timedelta(seconds=resolution.seconds * n)

    @classmethod
    def add_one_resolution(
        cls,
        date: datetime,
        resolution: ResolutionDetail,
    ) -> datetime:
        return date + timedelta(seconds=resolution.seconds)

    @classmethod
    def assert_candle_unity(cls, candles: List[Candle]) -> None:
        if len(candles) <= 1:
            return
        market = candles[0].market
        resolution = candles[0].resolution
        exchange = candles[0].exchange

        fail_value = None
        try:
            fail_value = next(
                (index, candle)
                for index, candle in enumerate(candles)
                if candle.market != market
                or candle.exchange != exchange
                or candle.resolution != resolution
            )
        except StopIteration:
            return
        index, candle = fail_value
        msg = "Candles need to be identical in exchange, market & resolution"
        mismatch = f"(candle #{index} is a mismatch: {candle})."
        raise AssertionError(f"{msg} {mismatch}")

    @classmethod
    def detect_missing_dates(
        cls,
        candle_dates: List[datetime],
        resolution: ResolutionDetail,
    ) -> List[datetime]:
        missing: List[datetime] = []

        if len(candle_dates) <= 1:
            return missing

        diff_ok = resolution.seconds
        for i, candle_date in enumerate(candle_dates):
            if i > 0:
                new = candle_dates[i - 1]
                old = candle_date
                if new - old != diff_ok:
                    missing_dates = cls.missing_dates_between(
                        old,
                        new,
                        resolution,
                    )
                    missing.extend(missing_dates)
        return missing

    @classmethod
    def amount_of_candles_in_range(
        cls,
        start_date: datetime,
        end_date: datetime,
        resolution: ResolutionDetail,
    ) -> int:
        assert start_date <= end_date, "start_date must be before end_date."

        range_seconds = (end_date - start_date).total_seconds()
        return int(range_seconds / resolution.seconds)

    @classmethod
    def amount_missing(
        cls,
        start_date: datetime,
        end_date: datetime,
        resolution: ResolutionDetail,
    ) -> int:
        return cls.amount_of_candles_in_range(start_date, end_date, resolution) - 1

    @classmethod
    def missing_dates_between(
        cls,
        start_date: datetime,
        end_date: datetime,
        resolution: ResolutionDetail,
    ) -> List[datetime]:
        missing_dates: List[datetime] = []
        current_date = start_date + timedelta(seconds=resolution.seconds)
        while current_date < end_date:
            missing_dates.append(current_date)
            current_date += timedelta(seconds=resolution.seconds)
        return missing_dates

    @classmethod
    def midpoint(cls, start: datetime, end: datetime) -> datetime:
        # Calculate the midpoint between two datetime objects.
        return start + (end - start) / 2



from datetime import datetime
from typing import List


from locast.candle.candle import Candle

from locast.candle.candle_utility import CandleUtility as cu
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_fetcher.candle_fetcher import CandleFetcher
from locast.candle_fetcher.dydx.api_fetcher.dydx_fetcher import DydxFetcher
from locast.candle_fetcher.exceptions import APIException
from locast.logging_functions import log_missing_candles, log_progress


class DydxCandleFetcher(CandleFetcher):
    def __init__(self, api_fetcher: DydxFetcher, log_progress: bool = False) -> None:
        self._log_progress = log_progress
        self._exchange = api_fetcher.exchange
        self._fetcher = api_fetcher

    @property
    def exchange(self) -> Exchange:
        return self._exchange

    @property
    def log_progress(self) -> bool:
        return self._log_progress

    @log_progress.setter
    def log_progress(self, value: bool) -> None:
        self._log_progress = value

    async def fetch_candles(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
        end_date: datetime,
    ) -> List[Candle]:
        """
        Fetches candles from exchange (derived from api_fetcher) and market within a given time range.

        Args:
            market (str): The market to fetch candles for.
            resolution (str): The resolution of the candles to fetch.
            start_date (datetime): The start date of the time range (the started_at value of the oldest candle in the range).
            end_date (datetime): The end date of the time range (the theoretical end date of the newest candle in the range).

        Returns:
            List[Candle]: A list of candles fetched from the specified exchange and market.

        Note: This function does not guarantee its returned candles to be reaching to the most recent existing candle on the exchange.
        It simply fetches candles making up the range between the provided start_date and end_date. If this takes longer than newer
        candles to be created on the exchange, those candles will not be fetched.
        """
        candles: List[Candle] = []
        temp_end_date = end_date
        total = cu.amount_of_candles_in_range(start_date, end_date, resolution)
        done = 0
        total_missing_candles: List[datetime] = []

        try:
            while (not candles) or candles[-1].started_at > start_date:
                candle_batch: List[Candle] = await self._fetcher.fetch(
                    market,
                    resolution,
                    start_date,
                    temp_end_date,
                )

                if not candle_batch:
                    break

                prev_oldest_date = candles[-1].started_at if candles else end_date

                if missing_in_batch := self._detect_missing_in_batch(
                    candle_batch,
                    prev_oldest_date,
                ):
                    total -= len(missing_in_batch)
                    total_missing_candles.extend(missing_in_batch)

                if self._log_progress:
                    done += len(candle_batch)
                    log_progress("🚛", market, "candles", "fetched", done, total)

                candles.extend(candle_batch)
                temp_end_date = candles[-1].started_at

            if total_missing_candles:
                log_missing_candles(
                    "🚨",
                    self._exchange,
                    market,
                    resolution,
                    total_missing_candles,
                )

        except Exception as e:
            raise APIException(self._exchange, market, resolution, e) from e

        return candles

    async def fetch_candles_up_to_now(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
    ) -> List[Candle]:
        """
        Fetches a cluster of candles, which is a group of chronologically sorted, uninterrupted candles ranging
        from a given start date up to the most recently finished candle.

        Args:
            exchange (Exchange): The exchange to fetch candles from.
            market (str): The market to fetch candles for.
            resolution (str): The resolution of the candles to fetch.
            start_date (datetime): The start date of the time range (the started_at value of the oldest candle in the range).

        Returns:
            List[Candle]: A list of candles from the start date up to the most recently finished candle.
        """
        candles: List[Candle] = []

        temp_start_date = start_date
        temp_norm_now = cu.normalized_now(resolution)

        # This is what we wait for: The newest candle (at index 0) to have started_at one resolution below NOW,
        # which only happens, if during fetch_candles a new candle started.
        temp_now_minus_res = cu.subtract_n_resolutions(temp_norm_now, resolution, 1)
        while (not candles) or candles[0].started_at < temp_now_minus_res:
            new_candles = await self.fetch_candles(
                market,
                resolution,
                temp_start_date,
                temp_norm_now,
            )

            # Sort repeated fetches to the front of gathered candles
            if new_candles:
                candles = new_candles + candles

            # Update input for next iteration
            next_started_at = cu.add_one_resolution(candles[0].started_at, resolution)

            temp_start_date = next_started_at
            temp_norm_now = cu.normalized_now(resolution)
            temp_now_minus_res = cu.subtract_n_resolutions(temp_norm_now, resolution, 1)

        return candles

    async def find_horizon(self, market: str, resolution: ResolutionDetail) -> datetime:
        now = cu.normalized_now(resolution)
        step = 1000
        upper_bound = now
        lower_bound = None

        # First: Exponential Back-Off Search to quickly reach over the horizon (datetime at which exchange does not provide data)
        while True:
            back_n = cu.subtract_n_resolutions(now, resolution, step)
            back_n_plus_resolution = cu.add_one_resolution(back_n, resolution)

            # Query single candle
            candle = await self._fetcher.fetch(
                market,
                resolution,
                back_n,
                back_n_plus_resolution,
            )
            if candle:
                # Candle still exists, continue back-off
                upper_bound = back_n  # Move the upper bound further back
                step *= 2
            else:
                # Reached over the horizon, candle does not exist
                lower_bound = back_n  # Set this as the lower bound
                break

        # Second: Binary Search within the determined range
        delta = upper_bound - lower_bound
        while delta.total_seconds() > resolution.seconds:
            mid_point = cu.norm_date(cu.midpoint(lower_bound, upper_bound), resolution)
            mid_point_plus_resolution = cu.add_one_resolution(mid_point, resolution)

            # Query single candle
            candle = await self._fetcher.fetch(
                market, resolution, mid_point, mid_point_plus_resolution
            )
            if candle:
                # Move the upper bound down (closer to the lower bound)
                upper_bound = mid_point
            else:
                # Move the lower bound up (closer to the upper bound)
                lower_bound = mid_point

            delta = upper_bound - lower_bound

        # Finally: Fetch to confirm the oldest available candle
        upper_bound_plus_resolution = cu.add_one_resolution(upper_bound, resolution)
        final_candle = await self._fetcher.fetch(
            market, resolution, upper_bound, upper_bound_plus_resolution
        )

        return final_candle[0].started_at

    def _detect_missing_in_batch(
        self,
        candle_batch: List[Candle],
        previous_last_candle: datetime,
    ) -> List[datetime]:
        dates_to_check = [candle.started_at for candle in candle_batch]

        # Insert last candle of previous batch to have overlap to cover candles missing in between batches
        dates_to_check.insert(0, previous_last_candle)

        return cu.detect_missing_dates(dates_to_check, candle_batch[0].resolution)



from typing import List, Protocol

from locast.candle.candle import Candle
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_storage.cluster_info import ClusterInfo


class CandleStorage(Protocol):
    async def store_candles(self, candles: List[Candle]) -> None: ...

    async def retrieve_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> List[Candle]: ...

    async def retrieve_newest_candles(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
        amount: int,
    ) -> List[Candle]: ...

    async def delete_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> None: ...

    async def get_cluster_info(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> ClusterInfo: ...



from locast.candle.exchange_resolution import (
    ResolutionDetail,
    Seconds,
    ExchangeResolution,
)


class DydxResolution(ExchangeResolution):
    ONE_DAY = ResolutionDetail(Seconds.ONE_DAY, "1DAY")
    FOUR_HOURS = ResolutionDetail(Seconds.FOUR_HOURS, "4HOURS")
    ONE_HOUR = ResolutionDetail(Seconds.ONE_HOUR, "1HOUR")
    THIRTY_MINUTES = ResolutionDetail(Seconds.THIRTY_MINUTES, "30MINS")
    FIFTEEN_MINUTES = ResolutionDetail(Seconds.FIFTEEN_MINUTES, "15MINS")
    FIVE_MINUTES = ResolutionDetail(Seconds.FIVE_MINUTES, "5MINS")
    ONE_MINUTE = ResolutionDetail(Seconds.ONE_MINUTE, "1MIN")



from datetime import datetime
from typing import Any, Dict, List

from dydx_v4_client.indexer.rest.indexer_client import IndexerClient  # type: ignore
from dydx_v4_client.network import TESTNET  # type: ignore


from locast.candle.candle import Candle

from locast.candle.dydx.dydx_candle_mapping import DydxV4CandleMapping
from locast.candle.exchange import Exchange
from locast.candle.exchange_candle_mapper import ExchangeCandleMapper
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_fetcher.dydx.api_fetcher.datetime_format import (
    datetime_to_dydx_iso_str,
)
from locast.candle_fetcher.dydx.api_fetcher.dydx_fetcher import DydxFetcher
from locast.candle_fetcher.exceptions import APIException


class DydxV4Fetcher(DydxFetcher):
    def __init__(
        self,
        client: IndexerClient = IndexerClient(TESTNET.rest_indexer),
    ) -> None:
        self._exchange = Exchange.DYDX_V4
        self._client = client
        self._mapper = ExchangeCandleMapper(DydxV4CandleMapping())

    @property
    def exchange(self) -> Exchange:
        return self._exchange

    async def fetch(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
        end_date: datetime,
    ) -> List[Candle]:
        try:
            response: Dict[
                str,
                Any,
            ] = await self._client.markets.get_perpetual_market_candles(  # type: ignore
                market=market,
                resolution=resolution.notation,
                from_iso=datetime_to_dydx_iso_str(start_date),
                to_iso=datetime_to_dydx_iso_str(end_date),
            )

        except Exception as e:
            raise APIException(self._exchange, market, resolution, e) from e

        return self._mapper.to_candles(response["candles"])



from dataclasses import dataclass

from locast.candle.candle import Candle


@dataclass
class ClusterInfo:
    newest_candle: Candle | None
    oldest_candle: Candle | None
    size: int
    is_uptodate: bool



from datetime import datetime
from typing import List

from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail


def log_progress(
    emoji: str,
    market: str,
    group_name: str,
    past_tense: str,
    amount_done: int,
    total: int,
) -> None:
    progress_message = f"{amount_done} of {total} {market}-{group_name} {past_tense}."
    if amount_done == total:
        print(f"\r{emoji} {progress_message} ✅", end="\n", flush=True)
    else:
        print(f"{emoji} {progress_message}", end="\r", flush=True)


def log_missing_candles(
    emoji: str,
    exchange: Exchange,
    market: str,
    resolution: ResolutionDetail,
    missing_candle_dates: List[datetime],
) -> None:
    n = len(missing_candle_dates)
    candles = "candles" if n > 1 else "candle"
    detail = f"{market}, {resolution.notation}"
    print(
        f"{emoji} Attention: {exchange.name} failed to deliver {n} {candles} for {detail} {emoji}"
    )
    for date in missing_candle_dates:
        print(f"    ❌ Candle missing: {date}.")


def log_start_date_shifted_to_horizon(
    emoji: str,
    exchange: Exchange,
    start_date: datetime,
    horizon: datetime,
) -> None:
    print(
        f"{emoji} Attention: Candle data on {exchange.name} only reaches back to {horizon}. To prevent over-fetching, the provided start date ({start_date}) will be shifted. {emoji}"
    )


def log_redundant_call(emoji: str, message: str) -> None:
    print(f"{emoji} {message}")



from datetime import datetime
from typing import List, Protocol

from locast.candle.candle import Candle
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail


class APIFetcher(Protocol):
    @property
    def exchange(self) -> Exchange: ...
    async def fetch(
        self,
        market: str,
        resolution: ResolutionDetail,
        start_date: datetime,
        end_date: datetime,
    ) -> List[Candle]: ...



from typing import List, Tuple
from sqlalchemy import Engine, and_, delete, literal

from sqlmodel import SQLModel, Session, asc, desc, select, func

from locast.candle.candle_utility import CandleUtility as cu
from locast.candle.candle import Candle
from locast.candle.exchange import Exchange
from locast.candle.exchange_resolution import ResolutionDetail
from locast.candle_storage.candle_storage import CandleStorage
from locast.candle_storage.cluster_info import ClusterInfo
from locast.candle_storage.database_candle_mapper import DatabaseCandleMapper
from locast.candle_storage.sql.sqlite_candle_mapping import SqliteCandleMapping
from locast.candle_storage.sql.tables import (
    SqliteCandle,
    SqliteExchange,
    SqliteMarket,
    SqliteResolution,
)

from locast.candle_storage.sql.table_utility import TableUtility as tu
from locast.logging_functions import log_progress


class SqliteCandleStorage(CandleStorage):
    def __init__(self, engine: Engine, log_progress: bool = False) -> None:
        self._log_progress = log_progress
        self._engine = engine
        SQLModel.metadata.create_all(self._engine)

    async def store_candles(self, candles: List[Candle]) -> None:
        # NOTE: Benchmarking showed stable, scalable results across different orders of magnitudes with a batch size of 5k
        await self._bulk_save_objects_batched(candles, 5000)

    async def retrieve_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> List[Candle]:
        with Session(self._engine) as session:
            if foreign_keys := self._look_up_foreign_keys(
                exchange,
                market,
                resolution,
                session,
            ):
                sqlite_exchange, sqlite_market, sqlite_resolution = foreign_keys

                stmnt = (
                    select(SqliteCandle)
                    .where(
                        (SqliteCandle.exchange_id == sqlite_exchange.id)
                        & (SqliteCandle.market_id == sqlite_market.id)
                        & (SqliteCandle.resolution_id == sqlite_resolution.id)
                    )
                    .order_by(desc(SqliteCandle.started_at))
                )

                results = session.exec(stmnt)
                return self._to_candles(list(results.all()))
            else:
                return []

    async def retrieve_newest_candles(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
        amount: int,
    ) -> List[Candle]:
        with Session(self._engine) as session:
            if foreign_keys := self._look_up_foreign_keys(
                exchange,
                market,
                resolution,
                session,
            ):
                sqlite_exchange, sqlite_market, sqlite_resolution = foreign_keys

                stmnt = (
                    select(SqliteCandle)
                    .where(
                        (SqliteCandle.exchange_id == sqlite_exchange.id)
                        & (SqliteCandle.market_id == sqlite_market.id)
                        & (SqliteCandle.resolution_id == sqlite_resolution.id)
                    )
                    .order_by(desc(SqliteCandle.started_at))
                    .limit(amount)
                )

                results = session.exec(stmnt)
                return self._to_candles(list(results.all()))
            else:
                return []

    async def delete_cluster(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> None:
        with Session(self._engine) as session:
            if foreign_keys := self._look_up_foreign_keys(
                exchange,
                market,
                resolution,
                session,
            ):
                sqlite_exchange, sqlite_market, sqlite_resolution = foreign_keys

                stmt = delete(SqliteCandle).where(
                    and_(
                        literal(sqlite_exchange.id) == SqliteCandle.exchange_id,
                        literal(sqlite_market.id) == SqliteCandle.market_id,
                        literal(sqlite_resolution.id) == SqliteCandle.resolution_id,
                    )
                )

                with self._engine.begin() as conn:
                    conn.execute(stmt)

                session.commit()

    async def get_cluster_info(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
    ) -> ClusterInfo:
        result = ClusterInfo(None, None, 0, False)
        with Session(self._engine) as session:
            if foreign_keys := self._look_up_foreign_keys(
                exchange,
                market,
                resolution,
                session,
            ):
                sqlite_exchange, sqlite_market, sqlite_resolution = foreign_keys

                if head_and_tail := self._query_head_and_tail(
                    sqlite_exchange,
                    sqlite_market,
                    sqlite_resolution,
                    session,
                ):
                    head, tail = head_and_tail
                    amount = self._query_amount(
                        sqlite_exchange,
                        sqlite_market,
                        sqlite_resolution,
                        session,
                    )
                    is_uptodate = cu.is_newest_valid_candle(head)
                    if head and amount:
                        result = ClusterInfo(head, tail, amount, is_uptodate)
        return result

    async def _bulk_save_objects_batched(
        self,
        candles: List[Candle],
        batch_size: int,
    ) -> None:
        done = 0
        total = len(candles)

        num_batches = (total + batch_size - 1) // batch_size  # Ceiling division

        for i in range(num_batches):
            # Get the current batch of candles
            batch = candles[i * batch_size : (i + 1) * batch_size]

            # Create a new session for this batch
            with Session(self._engine) as session:
                # Initialize mapper with the current session
                mapper = DatabaseCandleMapper(SqliteCandleMapping(session))

                # Perform bulk insert
                session.bulk_save_objects(
                    [mapper.to_database_candle(candle) for candle in batch]
                )
                session.commit()

                if self._log_progress:
                    done += len(batch)
                    market = batch[0].market
                    log_progress("📀", market, "candles", "stored", done, total)

    def _to_candles(self, database_candles: List[SqliteCandle]) -> List[Candle]:
        mapper = DatabaseCandleMapper(SqliteCandleMapping())
        return [mapper.to_candle(sqlite_candle) for sqlite_candle in database_candles]

    def _look_up_foreign_keys(
        self,
        exchange: Exchange,
        market: str,
        resolution: ResolutionDetail,
        session: Session,
    ) -> Tuple[SqliteExchange, SqliteMarket, SqliteResolution] | None:
        sqlite_exchange = tu.lookup_sqlite_exchange(exchange, session)
        sqlite_market = tu.lookup_sqlite_market(market, session)
        sqlite_resolution = tu.lookup_sqlite_resolution(resolution, session)

        if not (sqlite_exchange and sqlite_market and sqlite_resolution):
            return None

        return sqlite_exchange, sqlite_market, sqlite_resolution

    def _query_head_and_tail(
        self,
        sqlite_exchange: SqliteExchange,
        sqlite_market: SqliteMarket,
        sqlite_resolution: SqliteResolution,
        session: Session,
    ) -> Tuple[Candle, Candle] | None:
        head = self._query_head(
            sqlite_exchange,
            sqlite_market,
            sqlite_resolution,
            session,
        )

        tail = self._query_tail(
            sqlite_exchange,
            sqlite_market,
            sqlite_resolution,
            session,
        )

        return (head, tail) if head and tail else None

    def _query_head(
        self,
        sqlite_exchange: SqliteExchange,
        sqlite_market: SqliteMarket,
        sqlite_resolution: SqliteResolution,
        session: Session,
    ) -> Candle | None:
        head_statement = (
            select(SqliteCandle)
            .where(
                (SqliteCandle.exchange_id == sqlite_exchange.id)
                & (SqliteCandle.market_id == sqlite_market.id)
                & (SqliteCandle.resolution_id == sqlite_resolution.id)
            )
            .order_by(desc(SqliteCandle.started_at))
            .limit(1)
        )

        head: Candle | None = None

        if head_results := session.exec(head_statement).first():
            head = self._to_candles([head_results])[0]

        return head

    def _query_tail(
        self,
        sqlite_exchange: SqliteExchange,
        sqlite_market: SqliteMarket,
        sqlite_resolution: SqliteResolution,
        session: Session,
    ) -> Candle | None:
        tail_statement = (
            select(SqliteCandle)
            .where(
                (SqliteCandle.exchange_id == sqlite_exchange.id)
                & (SqliteCandle.market_id == sqlite_market.id)
                & (SqliteCandle.resolution_id == sqlite_resolution.id)
            )
            .order_by(asc(SqliteCandle.started_at))
            .limit(1)
        )

        tail: Candle | None = None

        if tail_results := session.exec(tail_statement).first():
            tail = self._to_candles([tail_results])[0]

        return tail

    def _query_amount(
        self,
        sqlite_exchange: SqliteExchange,
        sqlite_market: SqliteMarket,
        sqlite_resolution: SqliteResolution,
        session: Session,
    ) -> int | None:
        stmnt = select(SqliteCandle).where(
            (SqliteCandle.exchange_id == sqlite_exchange.id)
            & (SqliteCandle.market_id == sqlite_market.id)
            & (SqliteCandle.resolution_id == sqlite_resolution.id)
        )

        count_query = (
            stmnt.with_only_columns(func.count())
            .order_by(None)
            .select_from(stmnt.get_final_froms()[0])
        )

        return session.scalar(count_query)
